# Configure macros for Ocapi
# Patrice Dumas 2005 based on freetype2.m4 from Marcelo Magallon 2001-10-26, 
# based on gtk.m4 by Owen Taylor

# AC_CHECK_OCAPI([MINIMUM-VERSION [, ACTION-IF-FOUND [, ACTION-IF-NOT-FOUND]]])
# Test for Ocapi and define OCAPI_CFLAGS and OCAPI_LIBS.
# Check that the version is above MINIMUM-VERSION 
# use when linking with a c++ aware linker, with a c linker you may also
# need -lstdc++

# Example use:
# AC_CHECK_OCAPI([3.5.0],
#  [
#   LIBS="$LIBS $OCAPI_LIBS"
#   CPPFLAGS="$CPPFLAGS $OCAPI_CFLAGS"
#  ],
#  [ AC_MSG_ERROR([Cannot find Ocapi])
# ])


AC_DEFUN([AC_CHECK_OCAPI],
[
  AC_PATH_PROG([OCAPI_CONFIG], [ocapi-config], [no])
  ocapi_min_version=m4_if([$1], [], [1.0], [$1])
  AC_MSG_CHECKING([for Ocapi version >= $ocapi_min_version])
  ocapi_no=""
  if test "$OCAPI_CONFIG" = "no" ; then
     ocapi_no=yes
  else
     ocapi_config_major_version=`$OCAPI_CONFIG --version | sed 's/^libocapi \([[0-9]]\)*\.\([[0-9]]*\)$/\1/'`
     ocapi_config_minor_version=`$OCAPI_CONFIG --version | sed 's/^libocapi \([[0-9]]\)*\.\([[0-9]]*\)$/\2/'`
     ocapi_min_major_version=`echo $ocapi_min_version | sed 's/\([[0-9]]*\).\([[0-9]]*\)/\1/'`
     ocapi_min_minor_version=`echo $ocapi_min_version | sed 's/\([[0-9]]*\).\([[0-9]]*\)/\2/'`

     ocapi_config_is_lt=""
     if test $ocapi_config_major_version -lt $ocapi_min_major_version ; then
       ocapi_config_is_lt=yes
     else
       if test $ocapi_config_major_version -eq $ocapi_min_major_version ; then
         if test $ocapi_config_minor_version -lt $ocapi_min_minor_version ; then
           ocapi_config_is_lt=yes
         fi
       fi
     fi
    if test x$ocapi_config_is_lt = xyes ; then
       ocapi_no=yes
     else
       OCAPI_LIBS="`$OCAPI_CONFIG --libs`"
       OCAPI_CFLAGS="`$OCAPI_CONFIG --cflags`"
     fi
   fi
   if test x$ocapi_no = x ; then
     AC_MSG_RESULT([yes])
     m4_if([$2], [], [:], [$2])
   else
     AC_MSG_RESULT([no])
     if test "$OCAPI_CONFIG" = "no" ; then
     AC_MSG_NOTICE([The ocapi-config script could not be found.])
     else
       if test x$ocapi_config_is_lt = xyes ; then
         AC_MSG_NOTICE([the installed Ocapi library is too old.])
       fi
     fi
     OCAPI_LIBS=""
     OCAPI_CFLAGS=""
     m4_if([$3], [], [:], [$3])
   fi
   AC_SUBST([OCAPI_CFLAGS])
   AC_SUBST([OCAPI_LIBS])]) 

