// -*- mode: c++; c-basic-offset:4 -*-

// IDL-OPENDAP interface.
// Copyright (c) 2006 OPeNDAP, Inc.
// Author: Patrick West <mrpatrickwest@gmail.com> UCAR/HAO
//         Dan Holloway <d.holloway@opendap.org>
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
 
// Please read the full copyright statement in the file COPYING.
//

// IDL "LINKIMAGE" routines.
// 
// dods_connect()
//

#include "config_idl.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#ifndef WIN32
#include <strings.h>
#endif

// OCAPI include
#include "ocapi.h"
extern ConnectionParameters gDefaultConnectionParameters;
extern ConversionFormats gConversionFormats;
void (*gpvUpdateProgress)(int iBytesReceived, int iBytesTotal, char* sStatus) = NULL;

void setProgress( void (*vUpdateProgress)(int iBytesReceived, int iBytesTotal, char* sStatus) )
{
    gpvUpdateProgress = vUpdateProgress;
}

// IDL include.
#include <export.h>

// opendap_connect()
// Establish connection and return connect object and DDS object.
//    argv[0] - (Input ) url byte stream (null terminated).
//    argv[1] - (Input ) Constraint expression byte stream (null terminated).
//    argv[2] - (Output) Named IDL variable to recieve a pointer
//    argv[3] - (Input ) ACCEPT_DEFLATE flag.
//    argv[4] - (Input ) A flag indicating what kind of data request is:
//    data=0, dds=1, das=2, fulldata=3, fulldds=4, fulldas=5
//    returns -          Status (1=ok, 0=fail).
//
// In version 3.0.x of DODS the progress dialog is not supported. In version
// 3.1 or 3.2 we will reintroduce this with a new, more robust,
// implementation. In this code I have disabled the GUI flag option (which
// activated the progress dialog feature). 6/8/99 jhrg
//
IDL_VPTR
opendap_connect( int argc, IDL_VPTR argv[] )
{
  /*
    FILE *fp = fopen( "opendap_connect.out", "w" ) ;
    fprintf( fp, "opendap_connect entered\n" ) ;
    fflush( fp ) ;
  */
    assert( argc == 5 ) ;

    IDL_VPTR lReturn ;
    lReturn = IDL_Gettmp(); // Allocate memory for return variable.
    lReturn->type = IDL_TYP_ULONG;
    lReturn->value.ul = 0ul;

    unsigned long accept_deflate = 0 ;

    char *err = NULL ;
    eRESULT res = SUCCESS ;
    URL ocapi_url ;
    DAPObject* dap_obj = NULL ;

    if( argv[3]->value.ul == 0ul ) accept_deflate = 0 ;
    else accept_deflate = 1 ;

    lReturn = IDL_Gettmp() ; // Allocate memory for return variable.
    lReturn->type = IDL_TYP_MEMINT ;
    lReturn->value.memint = 0ul ;

    gDefaultConnectionParameters.sUserAgent = "IDL/Ocapi" ;
    char *url = (char *)argv[0]->value.arr->data ;
    char *ext = "" ;
    char *expr = (char *)argv[1]->value.arr->data ;
    /*fprintf( fp, "constraint_expression: %s\n", expr ) ;*/

    char *fullurl = 0 ;

    unsigned long connect_mode = argv[4]->value.ul ;

    // If the user is requesting a dds, the constraint expression needs
    // to be appended to the URL for the dds to be constrained.  
    // Otherwise the two can be kept seperate and the CE applied later
    // rph 06/04/2001.
    switch( connect_mode )
    {
	case 5:
	    {
		ocapi_url.eRequestClass = RC_DAS ;
	    }
	    break ;
	case 4:
	    {
		ocapi_url.eRequestClass = RC_DDS ;
	    }
	    break ;
	case 3:
	    {
		ocapi_url.eRequestClass = RC_Data ;
	    }
	    break ;
	case 2:
	    {
		ocapi_url.eRequestClass = RC_DAS ;
		ext = ".das" ;
	    }
	    break ;
	case 1:
	    {
		ocapi_url.eRequestClass = RC_DDS ;
		ext = ".dds" ;
	    }
	    break ;
	case 0:
	    {
		ocapi_url.eRequestClass = RC_Data ;
		ext = ".dods" ;
	    }
	    break ;
    }

    if( connect_mode == 1 && strlen( expr ) != 0 ) 
    {
	int fullsize = strlen( url ) + strlen( ext ) + strlen( expr ) ;
	fullurl = (char *)malloc( fullsize + 2 ) ;
	sprintf( fullurl, "%s%s?%s", url, ext, expr ) ;
    }
    else 
    {
	int fullsize;
	if ( strlen( expr ) != 0 ) {
	  fullsize = strlen( url ) + strlen( ext ) + strlen( expr );
	  fullurl = (char *)malloc( fullsize + 2 ) ;
	  sprintf( fullurl, "%s%s?%s", url, ext, expr ) ;
	}
	else {
	  fullsize = strlen( url ) + strlen( ext );
	  fullurl = (char *)malloc( fullsize + 1 ) ;
	  sprintf( fullurl, "%s%s", url, ext ) ;
	}
    }
    /*
      fprintf( fp, "connecting to %s\n", fullurl ) ;
      fflush( fp ) ;
    */

    // standard Url elements that don't need to change
    ocapi_url.sHost = NULL ;
    ocapi_url.sPath = NULL ;
    ocapi_url.sQuery = NULL ;

    // Go get a DAS object and go through it
    ocapi_url.sRawURL = fullurl ;
    ocapi_url.sCanonicalURL = fullurl ;

    /*
      fprintf( fp, "calling structure_Load_URL\n" ) ;
      fflush( fp ) ;
    */
    res = structure_Load_URL( &ocapi_url, &dap_obj, &err );
    /*
      fprintf( fp, "done calling structure_Load_URL: %d\n", res ) ;
      fflush( fp ) ;
    */
    if( res == FAILURE )
    {
      /*fprintf( fp, "connect failed\n" ) ;
	fflush( fp ) ;
	fclose( fp ) ;
      */
	free( fullurl ) ;
	if( err )
	{
	    printf( "%s\n", err ) ;
	}
	argv[2]->value.memint = (IDL_MEMINT)0UL ;
	return lReturn; // Connection failed.
    }

    if( dap_obj == 0 )
    {
      /*fprintf( fp, "null dap_obj\n" ) ;
	fflush( fp ) ;
	fclose( fp ) ;
      */
	free( fullurl ) ;
	argv[2]->value.memint = (IDL_MEMINT)0UL ;
	return lReturn; // Request_data failed.
    }

    if( ( connect_mode == 2 || connect_mode == 5 ) &&
        dap_obj->eStructureType != STRUCTURE_DAS )
    {
	char *myerr = "URL did not return an OPeNDAP DAS object" ;
	printf( "%s\n", myerr ) ;
	structure_Free( dap_obj, &err ) ;
	argv[2]->value.memint = (IDL_MEMINT)0UL ;
	return lReturn ;
    }

    if( ( connect_mode == 1 || connect_mode == 4 ) &&
        dap_obj->eStructureType != STRUCTURE_DDS )
    {
	char *myerr = "URL did not return an OPeNDAP DDS object" ;
	printf( "%s: returned %d\n", myerr, dap_obj->eStructureType ) ;
	structure_Free( dap_obj, &err ) ;
	argv[2]->value.memint = (IDL_MEMINT)0UL ;
	return lReturn ;
    }

    if( ( connect_mode == 0 || connect_mode == 3 ) &&
        dap_obj->eStructureType != STRUCTURE_DataDDS )
    {
      /*fprintf( fp, "Data DDS not returned\n" ) ;
	fflush( fp ) ;
	fclose( fp ) ;
      */
	char *myerr = "URL did not return an OPeNDAP Data DDS object" ;
	printf( "%s\n", myerr ) ;
	if( dap_obj->eStructureType == STRUCTURE_undefined )
	{
	    printf( "returned an undefined structure\n", myerr ) ;
	}
	if( dap_obj->eStructureType == STRUCTURE_DataDDS )
	{
	    printf( "returned a DataDDS\n", myerr ) ;
	}
	if( dap_obj->eStructureType == STRUCTURE_DAS )
	{
	    printf( "returned a DAS\n", myerr ) ;
	}
	if( dap_obj->eStructureType == STRUCTURE_DDS )
	{
	    printf( "returned a DDS\n", myerr ) ;
	}
	if( dap_obj->eStructureType == STRUCTURE_DDX )
	{
	    printf( "returned a DDX\n", myerr ) ;
	}
	structure_Free( dap_obj, &err ) ;
	argv[2]->value.memint = (IDL_MEMINT)0UL ;
	return lReturn ;
    }

    /*
      fprintf( fp, "getting the root node\n" ) ;
      fflush( fp ) ;
    */
    DAPNode *root_node = dap_obj->nodeRoot ;

    argv[2]->value.memint = (IDL_MEMINT)root_node ;

    /*
      fprintf( fp, "freeing the url\n" ) ;
      fflush( fp ) ;
    */
    // free allocated memory
    free( fullurl ) ;

    // Must have succeeded
    lReturn->value.ul = 1ul;

    /*
      fprintf( fp, "returning\n" ) ;
      fflush( fp ) ;
      fclose( fp ) ;
    */
    return lReturn ;
}

// opendap_node_info()
// Get information about a node.
//    argv[0] - (Input ) node pointer.
//    argv[1] - (Output) Named IDL variable to receive node name
//    argv[2] - (Output) Named IDL variable to receive node type
//    argv[3] - (Output) Named IDL variable to receive number of sub nodes
//    argv[4] - (Output) Named IDL variable to receive whether node has content
//    argv[5] - (Output) Named IDL variable to receive whether node has attrs
//    argv[6] - (Output) Named IDL variable to receive node content ptr
//
void
opendap_node_info( int argc, IDL_VPTR argv[] )
{
    DAPNode *node = (DAPNode *) argv[0]->value.memint ;
    if( !node )
    {
	return ;
    }

    char *node_name = node->sName ;
    IDL_VPTR name_arr ;
    name_arr = IDL_Gettmp() ;
    name_arr->type = IDL_TYP_STRING ;
    name_arr->value.str.slen = strlen( node_name ) ;
    name_arr->value.str.stype = 0 ;
    name_arr->value.str.s = (char *)node_name ;
    IDL_VarCopy( name_arr, argv[1] ) ;

    IDL_VPTR node_type ;
    node_type = IDL_Gettmp() ;
    node_type->type = IDL_TYP_ULONG ;
    node_type->value.ul = (IDL_ULONG)node->eType ;
    IDL_VarCopy( node_type, argv[2] ) ;

    IDL_VPTR num_subnodes ;
    num_subnodes = IDL_Gettmp() ;
    num_subnodes->type = IDL_TYP_ULONG ;
    num_subnodes->value.ul = (IDL_ULONG)node->iSubNodeCount ;
    IDL_VarCopy( num_subnodes, argv[3] ) ;

    IDL_VPTR has_content ;
    has_content = IDL_Gettmp() ;
    has_content->type = IDL_TYP_ULONG ;
    if( node->nodeContent )
    {
	has_content->value.ul = (IDL_ULONG)1 ;
    }
    else
    {
	has_content->value.ul = (IDL_ULONG)0 ;
    }
    IDL_VarCopy( has_content, argv[4] ) ;

    IDL_VPTR has_attrs ;
    has_attrs = IDL_Gettmp() ;
    has_attrs->type = IDL_TYP_ULONG ;
    if( node->attributes )
    {
	has_attrs->value.ul = (IDL_ULONG)1 ;
    }
    else
    {
	has_attrs->value.ul = (IDL_ULONG)0 ;
    }
    IDL_VarCopy( has_attrs, argv[5] ) ;

    IDL_VPTR content ;
    content = IDL_Gettmp() ;
    content->type = IDL_TYP_MEMINT ;
    content->value.memint = (IDL_MEMINT)0UL ;
    if( node->nodeContent )
    {
	content->value.memint = (IDL_MEMINT)node->nodeContent ;
    }
    IDL_VarCopy( content, argv[6] ) ;
}

// opendap_get_subnode()
// Get the specified sub node of the specified node
//    argv[0] - (Input ) node pointer.
//    argv[1] - (Input ) number of subnode to get
//    argv[2] - (Output) Named IDL variable to receive subnode
//    argv[3] - (Output) Named IDL variable to receive subnode name
//
void
opendap_get_subnode( int argc, IDL_VPTR argv[] )
{
    IDL_VPTR sub_node ;
    sub_node = IDL_Gettmp() ;
    sub_node->type = IDL_TYP_MEMINT ;
    sub_node->value.memint = (IDL_MEMINT)0UL ;

    DAPNode *node = (DAPNode *) argv[0]->value.memint ;
    if( !node )
    {
	IDL_VarCopy( sub_node, argv[2] ) ;
	return ;
    }

    unsigned int sub_node_num = argv[1]->value.ul ;
    if( sub_node_num >= node->iSubNodeCount )
    {
	IDL_VarCopy( sub_node, argv[2] ) ;
	return ;
    }

    sub_node->value.memint = (IDL_MEMINT)node->nodeSub[sub_node_num] ;
    IDL_VarCopy( sub_node, argv[2] ) ;

    char *node_name = node->nodeSub[sub_node_num]->sName ;
    IDL_VPTR name_arr ;
    name_arr = IDL_Gettmp() ;
    name_arr->type = IDL_TYP_STRING ;
    name_arr->value.str.slen = strlen( node_name ) ;
    name_arr->value.str.stype = 0 ;
    name_arr->value.str.s = (char *)node_name ;
    IDL_VarCopy( name_arr, argv[3] ) ;
}

// opendap_get_attribute_table()
// Get the attribute table for the specified node
//    argv[0] - (Input ) node pointer.
//    argv[1] - (Output) Named IDL variable to receive attribute table
//    argv[2] - (Output) Named IDL variable to receive num of attrs in table
//
void
opendap_get_attribute_table( int argc, IDL_VPTR argv[] )
{
    IDL_VPTR attr_table ;
    attr_table = IDL_Gettmp() ;
    attr_table->type = IDL_TYP_MEMINT ;
    attr_table->value.memint = (IDL_MEMINT)0UL ;

    IDL_VPTR num_attrs ;
    num_attrs = IDL_Gettmp() ;
    num_attrs->type = IDL_TYP_ULONG ;
    num_attrs->value.ul = (IDL_ULONG)0 ;

    DAPNode *node = (DAPNode *) argv[0]->value.memint ;
    if( !node )
    {
	IDL_VarCopy( attr_table, argv[1] ) ;
	IDL_VarCopy( num_attrs, argv[2] ) ;
	return ;
    }

    if( !node->attributes )
    {
	IDL_VarCopy( attr_table, argv[1] ) ;
	IDL_VarCopy( num_attrs, argv[2] ) ;
	return ;
    }

    attr_table->value.memint = (IDL_MEMINT)node->attributes ;
    IDL_VarCopy( attr_table, argv[1] ) ;

    num_attrs->value.ul = (IDL_ULONG)node->attributes->ctAttributes ;
    IDL_VarCopy( num_attrs, argv[2] ) ;
}

// opendap_get_attribute()
// Get the specified attribute from the specified attribute table
//    argv[0] - (Input ) attribute table pointer.
//    argv[1] - (Input ) number of attribute to get
//    argv[2] - (Output) Named IDL variable to receive attribute
//
void
opendap_get_attribute( int argc, IDL_VPTR argv[] )
{
    IDL_VPTR attr ;
    attr = IDL_Gettmp() ;
    attr->type = IDL_TYP_MEMINT ;
    attr->value.memint = (IDL_MEMINT)0UL ;

    DAPAttributeTable *attr_table = (DAPAttributeTable *) argv[0]->value.memint ;
    if( !attr_table )
    {
	IDL_VarCopy( attr, argv[2] ) ;
	return ;
    }

    unsigned int attr_num = argv[1]->value.ul ;
    if( attr_num >= attr_table->ctAttributes )
    {
	IDL_VarCopy( attr, argv[2] ) ;
	return ;
    }

    attr->value.memint = (IDL_MEMINT)attr_table->pAttributeList[attr_num] ;
    IDL_VarCopy( attr, argv[2] ) ;
}

// opendap_get_attribute_info()
// Get the specified attribute information
//    argv[0] - (Input ) attribute pointer.
//    argv[1] - (Output) Named IDL variable to receive attribute name
//    argv[2] - (Output) Named IDL variable to receive attribute type
//    argv[3] - (Output) Named IDL variable to receive num values in attribute
//
void
opendap_get_attribute_info( int argc, IDL_VPTR argv[] )
{
    IDL_VPTR name_arr;
    name_arr = IDL_Gettmp() ;
    name_arr->type = IDL_TYP_STRING ;
    name_arr->value.str.slen = 0 ;
    name_arr->value.str.stype = 0 ;
    name_arr->value.str.s = (char *)"" ;

    IDL_VPTR attr_type ;
    attr_type = IDL_Gettmp() ;
    attr_type->type = IDL_TYP_ULONG ;
    attr_type->value.ul = (IDL_ULONG)0 ;

    IDL_VPTR num_values ;
    num_values = IDL_Gettmp() ;
    num_values->type = IDL_TYP_ULONG ;
    num_values->value.ul = (IDL_ULONG)0 ;

    DAPAttribute *attr = (DAPAttribute *) argv[0]->value.memint ;
    if( !attr )
    {
	IDL_VarCopy( name_arr, argv[1] ) ;
	IDL_VarCopy( attr_type, argv[2] ) ;
	IDL_VarCopy( num_values, argv[3] ) ;
	return ;
    }

    name_arr->value.str.slen = strlen( attr->sName ) ;
    name_arr->value.str.s = (char *)attr->sName ;
    IDL_VarCopy( name_arr, argv[1] ) ;

    attr_type->value.ul = (IDL_ULONG)attr->eType ;
    IDL_VarCopy( attr_type, argv[2] ) ;

    num_values->value.ul = (IDL_ULONG)attr->ctValues ;
    IDL_VarCopy( num_values, argv[3] ) ;
}

// opendap_get_attribute_value()
// Get the specified attribute information
//    argv[0] - (Input ) attribute pointer.
//    argv[1] - (Input ) attribute type.
//    argv[2] - (Input ) value num.
//    argv[3] - (Output) Named IDL variable to receive attribute value
//
void
opendap_get_attribute_value( int argc, IDL_VPTR argv[] )
{
    DAPAttribute *attr = (DAPAttribute *)argv[0]->value.memint ;
    if( !attr )
    {
	return ;
    }

    unsigned int attr_type = argv[1]->value.ul ;
    unsigned int value_num = argv[2]->value.ul ;
    switch( (eATTRIBUTE_TYPE)attr_type )
    {
	case ATTRIBUTE_Byte:
	    {
		UCHAR *val = (UCHAR *)attr->pValueList[value_num] ;
		IDL_VPTR val_arr;
		val_arr = IDL_Gettmp() ;
		val_arr->type = IDL_TYP_BYTE ;
		val_arr->value.c = (UCHAR)*val ;
		IDL_VarCopy( val_arr, argv[3] ) ;
	    }
	    break ;
	case ATTRIBUTE_Int16:
	    {
		int *val = (int *)attr->pValueList[value_num] ;
		IDL_VPTR val_arr;
		val_arr = IDL_Gettmp() ;
		val_arr->type = IDL_TYP_INT ;
		val_arr->value.i = *val ;
		IDL_VarCopy( val_arr, argv[3] ) ;
	    }
	    break ;
	case ATTRIBUTE_UInt16:
	    {
		unsigned int *val =
		    (unsigned int *)attr->pValueList[value_num] ;
		IDL_VPTR val_arr;
		val_arr = IDL_Gettmp() ;
		val_arr->type = IDL_TYP_UINT ;
		val_arr->value.ui = *val ;
		IDL_VarCopy( val_arr, argv[3] ) ;
	    }
	    break ;
	case ATTRIBUTE_Int32:
	    {
		long *val = (long *)attr->pValueList[value_num] ;
		IDL_VPTR val_arr;
		val_arr = IDL_Gettmp() ;
		val_arr->type = IDL_TYP_LONG ;
		val_arr->value.l = *val ;
		IDL_VarCopy( val_arr, argv[3] ) ;
	    }
	    break ;
	case ATTRIBUTE_UInt32:
	    {
		unsigned long *val =
		    (unsigned long *)attr->pValueList[value_num] ;
		IDL_VPTR val_arr;
		val_arr = IDL_Gettmp() ;
		val_arr->type = IDL_TYP_ULONG ;
		val_arr->value.ul = *val ;
		IDL_VarCopy( val_arr, argv[3] ) ;
	    }
	    break ;
	case ATTRIBUTE_Float32:
	    {
		float *val = (float *)attr->pValueList[value_num] ;
		IDL_VPTR val_arr;
		val_arr = IDL_Gettmp() ;
		val_arr->type = IDL_TYP_FLOAT ;
		val_arr->value.f = *val ;
		IDL_VarCopy( val_arr, argv[3] ) ;
	    }
	    break ;
	case ATTRIBUTE_Float64:
	    {
		double *val = (double *)attr->pValueList[value_num] ;
		IDL_VPTR val_arr;
		val_arr = IDL_Gettmp() ;
		val_arr->type = IDL_TYP_DOUBLE ;
		val_arr->value.d = *val ;
		IDL_VarCopy( val_arr, argv[3] ) ;
	    }
	    break ;
	case ATTRIBUTE_String:
	    {
		char *val = (char *)attr->pValueList[value_num] ;
		IDL_VPTR val_arr;
		val_arr = IDL_Gettmp() ;
		val_arr->type = IDL_TYP_STRING ;
		val_arr->value.str.slen = strlen( val ) ;
		val_arr->value.str.stype = 0 ;
		val_arr->value.str.s = (char *)val ;
		IDL_VarCopy( val_arr, argv[3] ) ;
	    }
	    break ;
	default:
	    {
	    }
	    break ;
    }
}

// opendap_get_byte()
// Get the value of the node as byte
//    argv[0] - (Input ) node pointer.
//    argv[1] - (Output) Named IDL variable to receive data value
//    argv[2] - (Input ) get_data, is this a dds or a datadds
//    returns - Status (1=ok, 0=fail)
//
IDL_VPTR
opendap_get_byte( int argc, IDL_VPTR argv[] )
{
    IDL_VPTR lReturn ;
    lReturn = IDL_Gettmp(); // Allocate memory for return variable.
    lReturn->type = IDL_TYP_ULONG;
    lReturn->value.ul = 0ul;

    char *arr_ptr ;
    IDL_MEMINT dim = 1 ;
    IDL_VPTR arr_data ;
    arr_ptr = IDL_MakeTempArray( IDL_TYP_BYTE, (int)1, 
				 (IDL_MEMINT *)&dim, IDL_BARR_INI_ZERO,
				 &arr_data ) ;
    IDL_VarCopy( arr_data, argv[1] ) ;

    unsigned long get_data = argv[2]->value.ul ;
    if( get_data )
    {
	void *content = (void *)argv[0]->value.memint ;
	if( !content )
	{
	    return lReturn ;
	}
	memcpy( argv[1]->value.arr->data, content, 1 ) ;
    }
    else
    {
	memset( argv[1]->value.arr->data, 0, 1 ) ;
    }

    lReturn->value.ul = 1ul;
    return lReturn ;
}

// opendap_get_int16()
// Get the value of the node as int16
//    argv[0] - (Input ) vontent pointer.
//    argv[1] - (Output) Named IDL variable to receive data value
//    argv[2] - (Input ) get_data, is this a dds or a datadds
//    returns - Status (1=ok, 0=fail)
//
IDL_VPTR
opendap_get_int16( int argc, IDL_VPTR argv[] )
{
    IDL_VPTR lReturn ;
    lReturn = IDL_Gettmp(); // Allocate memory for return variable.
    lReturn->type = IDL_TYP_ULONG;
    lReturn->value.ul = 0ul;

    char *arr_ptr ;
    IDL_MEMINT dim = 1 ;
    IDL_VPTR arr_data ;
    arr_ptr = IDL_MakeTempArray( IDL_TYP_INT, (int)1,
				 (IDL_MEMINT *)&dim, IDL_BARR_INI_ZERO,
				 &arr_data ) ;
    IDL_VarCopy( arr_data, argv[1] ) ;

    unsigned long get_data = argv[2]->value.ul ;
    if( get_data )
    {
	void *content = (void *)argv[0]->value.memint ;
	if( !content )
	{
	    return lReturn ;
	}
	memcpy( argv[1]->value.arr->data, content, 2 ) ;
    }
    else
    {
	memset( argv[1]->value.arr->data, 0, 2 ) ;
    }

    lReturn->value.ul = 1ul;
    return lReturn ;
}

// opendap_get_uint16()
// Get the value of the node as uint16
//    argv[0] - (Input ) content pointer.
//    argv[1] - (Output) Named IDL variable to receive data value
//    argv[2] - (Input ) get_data, is this a dds or a datadds
//    returns - Status (1=ok, 0=fail)
//
IDL_VPTR
opendap_get_uint16( int argc, IDL_VPTR argv[] )
{
    IDL_VPTR lReturn ;
    lReturn = IDL_Gettmp(); // Allocate memory for return variable.
    lReturn->type = IDL_TYP_ULONG;
    lReturn->value.ul = 0ul;

    char *arr_ptr ;
    IDL_MEMINT dim = 1 ;
    IDL_VPTR arr_data ;
    arr_ptr = IDL_MakeTempArray( IDL_TYP_UINT, (int)1,
				 (IDL_MEMINT *)&dim, IDL_BARR_INI_ZERO,
				 &arr_data ) ;
    IDL_VarCopy( arr_data, argv[1] ) ;

    unsigned long get_data = argv[2]->value.ul ;
    if( get_data )
    {
	void *content = (void *)argv[0]->value.memint ;
	if( !content )
	{
	    return lReturn ;
	}
	memcpy( argv[1]->value.arr->data, content, 2 ) ;
    }
    else
    {
	memset( argv[1]->value.arr->data, 0, 2 ) ;
    }

    lReturn->value.ul = 1ul;
    return lReturn ;
}

// opendap_get_int32()
// Get the value of the node as int32
//    argv[0] - (Input ) content pointer.
//    argv[1] - (Output) Named IDL variable to receive data value
//    argv[2] - (Input ) get_data, is this a dds or a datadds
//    returns - Status (1=ok, 0=fail)
//
IDL_VPTR
opendap_get_int32( int argc, IDL_VPTR argv[] )
{
    IDL_VPTR lReturn ;
    lReturn = IDL_Gettmp(); // Allocate memory for return variable.
    lReturn->type = IDL_TYP_ULONG;
    lReturn->value.ul = 0ul;

    char *arr_ptr ;
    IDL_MEMINT dim = 1 ;
    IDL_VPTR arr_data ;
    arr_ptr = IDL_MakeTempArray( IDL_TYP_LONG, (int)1,
				 (IDL_MEMINT *)&dim, IDL_BARR_INI_ZERO,
				 &arr_data ) ;
    IDL_VarCopy( arr_data, argv[1] ) ;

    unsigned long get_data = argv[2]->value.ul ;
    if( get_data )
    {
	void *content = (void *)argv[0]->value.memint ;
	if( !content )
	{
	    return lReturn ;
	}
	memcpy( argv[1]->value.arr->data, content, 4 ) ;
    }
    else
    {
	memset( argv[1]->value.arr->data, 0, 4 ) ;
    }

    lReturn->value.ul = 1ul;
    return lReturn ;
}

// opendap_get_uint32()
// Get the value of the node as uint32
//    argv[0] - (Input ) vontent pointer.
//    argv[1] - (Output) Named IDL variable to receive data value
//    argv[2] - (Input ) get_data, is this a dds or a datadds
//    returns - Status (1=ok, 0=fail)
//
IDL_VPTR
opendap_get_uint32( int argc, IDL_VPTR argv[] )
{
    IDL_VPTR lReturn ;
    lReturn = IDL_Gettmp(); // Allocate memory for return variable.
    lReturn->type = IDL_TYP_ULONG;
    lReturn->value.ul = 0ul;

    char *arr_ptr ;
    IDL_MEMINT dim = 1 ;
    IDL_VPTR arr_data ;
    arr_ptr = IDL_MakeTempArray( IDL_TYP_ULONG, (int)1,
				 (IDL_MEMINT *)&dim, IDL_BARR_INI_ZERO,
				 &arr_data ) ;
    IDL_VarCopy( arr_data, argv[1] ) ;

    unsigned long get_data = argv[2]->value.ul ;
    if( get_data )
    {
	void *content = (void *)argv[0]->value.memint ;
	if( !content )
	{
	    return lReturn ;
	}
	memcpy( argv[1]->value.arr->data, content, 4 ) ;
    }
    else
    {
	memset( argv[1]->value.arr->data, 0, 4 ) ;
    }

    lReturn->value.ul = 1ul;
    return lReturn ;
}

// opendap_get_float32()
// Get the value of the node as float32
//    argv[0] - (Input ) content pointer.
//    argv[1] - (Output) Named IDL variable to receive data value
//    argv[2] - (Input ) get_data, is this a dds or a datadds
//    returns - Status (1=ok, 0=fail)
//
IDL_VPTR
opendap_get_float32( int argc, IDL_VPTR argv[] )
{
    IDL_VPTR lReturn ;
    lReturn = IDL_Gettmp(); // Allocate memory for return variable.
    lReturn->type = IDL_TYP_ULONG;
    lReturn->value.ul = 0ul;

    char *arr_ptr ;
    IDL_MEMINT dim = 1 ;
    IDL_VPTR arr_data ;
    arr_ptr = IDL_MakeTempArray( IDL_TYP_FLOAT, (int)1,
				 (IDL_MEMINT *)&dim, IDL_BARR_INI_ZERO,
				 &arr_data ) ;
    IDL_VarCopy( arr_data, argv[1] ) ;

    unsigned long get_data = argv[2]->value.ul ;
    if( get_data )
    {
	void *content = (void *)argv[0]->value.memint ;
	if( !content )
	{
	    return lReturn ;
	}
	memcpy( argv[1]->value.arr->data, content, 4 ) ;
    }
    else
    {
	memset( argv[1]->value.arr->data, 0, 4 ) ;
    }

    lReturn->value.ul = 1ul;
    return lReturn ;
}

// opendap_get_float64()
// Get the value of the node as float64
//    argv[0] - (Input ) content pointer.
//    argv[1] - (Output) Named IDL variable to receive data value
//    argv[2] - (Input ) get_data, is this a dds or a datadds
//    returns - Status (1=ok, 0=fail)
//
IDL_VPTR
opendap_get_float64( int argc, IDL_VPTR argv[] )
{
    IDL_VPTR lReturn ;
    lReturn = IDL_Gettmp(); // Allocate memory for return variable.
    lReturn->type = IDL_TYP_ULONG;
    lReturn->value.ul = 0ul;

    char *arr_ptr ;
    IDL_MEMINT dim = 1 ;
    IDL_VPTR arr_data ;
    arr_ptr = IDL_MakeTempArray( IDL_TYP_DOUBLE, (int)1,
				 (IDL_MEMINT *)&dim, IDL_BARR_INI_ZERO,
				 &arr_data ) ;
    IDL_VarCopy( arr_data, argv[1] ) ;

    unsigned long get_data = argv[2]->value.ul ;
    if( get_data )
    {
	void *content = (void *)argv[0]->value.memint ;
	if( !content )
	{
	    return lReturn ;
	}
	memcpy( argv[1]->value.arr->data, content, 8 ) ;
    }
    else
    {
	memset( argv[1]->value.arr->data, 0, 8 ) ;
    }

    lReturn->value.ul = 1ul;
    return lReturn ;
}

// opendap_get_string()
// Get the value of the node as string
//    argv[0] - (Input ) content pointer.
//    argv[1] - (Output) Named IDL variable to receive data value
//    argv[2] - (Input ) get_data, is this a dds or a datadds
//    returns - Status (1=ok, 0=fail)
//
IDL_VPTR
opendap_get_string( int argc, IDL_VPTR argv[] )
{
    IDL_VPTR lReturn ;
    lReturn = IDL_Gettmp() ; // Allocate memory for return variable.
    lReturn->type = IDL_TYP_ULONG ;
    lReturn->value.ul = 0ul ;

    IDL_VPTR temp_string ;
    IDL_MEMINT dim = 0 ;

    char *str = 0 ;
    unsigned long get_data = argv[2]->value.ul ;
    if( get_data )
    {
	void *vcontent = (void *)argv[0]->value.memint ;
	if( !vcontent )
	{
	    return lReturn ;
	}
	NodeContent_String *content = (NodeContent_String *)vcontent ;
	str = (char *)malloc( content->iLength + 1 ) ;
	strncpy( str, content->sContent, content->iLength ) ;
	str[content->iLength] = '\0' ;
	dim = content->iLength + 1 ;
    }
    else
    {
	str = (char *)malloc( strlen( "" ) + 1 ) ;
	strcpy( str, "" ) ;
    }

    temp_string = IDL_Gettmp() ;
    temp_string->type = IDL_TYP_STRING ;
    temp_string->value.str.slen = strlen( str ) ;
    temp_string->value.str.stype = 0 ;
    temp_string->value.str.s = str ;
    IDL_VarCopy(temp_string, argv[1]);

    lReturn->value.ul = 1ul;

    return lReturn ;
}

// opendap_get_array()
// Get the value of the node as array
//    argv[0] - (Input ) content pointer.
//    argv[1] - (Output) Named IDL variable to receive array data
//    argv[2] - (Output) Named IDL variable to receive array name
//    argv[3] - (Output) Named IDL variable to receive dimension names
//    argv[4] - (Input ) get_data, is this a dds or a datadds
//    returns - Status (1=ok, 0=fail)
//
IDL_VPTR
opendap_get_array( int argc, IDL_VPTR argv[] )
{
    IDL_VPTR lReturn ;
    lReturn = IDL_Gettmp() ; // Allocate memory for return variable.
    lReturn->type = IDL_TYP_ULONG ;
    lReturn->value.ul = 0ul ;

    void *vcontent = (void *)argv[0]->value.memint ;
    if( !vcontent )
    {
	return lReturn ;
    }

    NodeContent_Array *content = (NodeContent_Array *)vcontent ;

    char *arr_name = (char *)malloc( strlen( content->sArrayName ) + 1 ) ;
    strcpy( arr_name, content->sArrayName ) ;
    IDL_VPTR arr_name_ptr ;
    arr_name_ptr = IDL_Gettmp() ;
    arr_name_ptr->type = IDL_TYP_STRING ;
    arr_name_ptr->value.str.stype = 0 ;
    arr_name_ptr->value.str.slen = strlen( arr_name ) ;
    arr_name_ptr->value.str.s = arr_name ;
    IDL_VarCopy( arr_name_ptr, argv[2] ) ;

    char *arr_ptr = 0 ;
    IDL_VPTR arr_data ;
    char *dim_name = 0 ;
    unsigned int dim_name_len = 0 ;
    UCHAR *uchar_p ;
    unsigned int offset = 0 ;
    unsigned int num_dims = content->ctDimensions ;
    unsigned int i_num_dims = 0 ;
    IDL_MEMINT dims[12] ;
    if( num_dims < 12 )
    {
	for( i_num_dims = 0; i_num_dims < num_dims; i_num_dims++ )
	{
	    dim_name = content->sDimensionName1[i_num_dims+1] ;
	    dim_name_len = 0 ;
	    if( dim_name ) dim_name_len = strlen( dim_name ) ;
	    if( dim_name_len > 64 ) dim_name_len = 64 ;
	    // Handle case where dimension is not a named dimension;
	    // Create a container for the dimension with a 'space' char in it.
	    if( dim_name_len == 0 )
	    {
		offset = 65 * i_num_dims ;
		uchar_p = (UCHAR *) &argv[3]->value.arr->data[0] + offset ;
		*uchar_p = ' ' ;
	    }
	    else
	    {
		unsigned int i_len = 0 ;
		for( i_len = 0; i_len < dim_name_len; i_len++ )
		{
		    offset = ( 65 * i_num_dims ) + i_len ;
		    uchar_p = (UCHAR *)&argv[3]->value.arr->data[0] + offset ;
		    *uchar_p = (UCHAR)dim_name[i_len];
		}
	    }
	    // IDL arrays have their dimensions in the reverse order from
	    // C/C++ arrays, so we have to declare the array dimensions 
	    // in reverse.  06/19/00  rph.
	    // dims[i] = var_p->dimension_size(q);
	    dims[num_dims - i_num_dims - 1] = content->iDimensionLengthArray1[i_num_dims+1] ;
	}
	unsigned int n_elements = 1;
	for( i_num_dims = 0; i_num_dims < num_dims; i_num_dims++ )
	{
	    n_elements *= dims[i_num_dims];
	}

	eARRAY_TYPE array_type = content->eType ;
	void *arr_vals = content->array ;
	unsigned long get_data = argv[4]->value.ul ;

	switch( array_type )
	{
	    case ARRAY_Byte:
	    {
		arr_ptr = IDL_MakeTempArray( IDL_TYP_BYTE, (int)num_dims, 
					     (IDL_MEMINT *)dims,
					     IDL_BARR_INI_NOP,
					     &arr_data ) ;
		if( get_data )
		{
		    memcpy( arr_data->value.arr->data,
		            arr_vals, n_elements * 1 ) ;
		}
		else 
		{
		    memset( arr_data->value.arr->data, 0, n_elements * 1 ) ;
		}

		IDL_VarCopy(arr_data, argv[1]);
	    }
	    break;

	    case ARRAY_Int16:
	    {
		arr_ptr = IDL_MakeTempArray( IDL_TYP_INT, (int)num_dims,
					     (IDL_MEMINT *)dims,
					     IDL_BARR_INI_NOP,
					     &arr_data ) ;
		if( get_data )
		{
		    memcpy( arr_data->value.arr->data,
		            arr_vals, n_elements * 2 ) ;
		}
		else 
		{
		    memset( arr_data->value.arr->data, 0, n_elements * 2 ) ;
		}

		IDL_VarCopy(arr_data, argv[1]);
	    }
	    break;
	    case ARRAY_UInt16:
	    {
		arr_ptr = IDL_MakeTempArray( IDL_TYP_UINT, (int)num_dims,
					     (IDL_MEMINT *)dims,
					     IDL_BARR_INI_NOP, &arr_data ) ;

		if( get_data )
		{
		    memcpy( arr_data->value.arr->data,
		            arr_vals, n_elements * 2 ) ;
		}
		else 
		{
		    memset( arr_data->value.arr->data, 0, n_elements * 2 ) ;
		}

		IDL_VarCopy(arr_data, argv[1]);
	    }
	    break;
	    case ARRAY_Int32:
	    {
		arr_ptr = IDL_MakeTempArray( IDL_TYP_LONG, (int)num_dims,
					     (IDL_MEMINT *)dims,
					     IDL_BARR_INI_NOP, &arr_data ) ;
		if( get_data )
		{
		    memcpy( arr_data->value.arr->data,
		            arr_vals, n_elements * 4 ) ;
		}
		else 
		{
		    memset( arr_data->value.arr->data, 0, n_elements * 4 ) ;
		}

		IDL_VarCopy( arr_data, argv[1] ) ;
	    }
	    break;
	    case ARRAY_UInt32:
	    {
		arr_ptr = IDL_MakeTempArray( IDL_TYP_ULONG, (int)num_dims,
					     (IDL_MEMINT *)dims,
					     IDL_BARR_INI_NOP, &arr_data ) ;
		if( get_data )
		{
		    memcpy( arr_data->value.arr->data,
		            arr_vals, n_elements * 4 ) ;
		}
		else 
		{
		    memset( arr_data->value.arr->data, 0, n_elements * 4 ) ;
		}

		IDL_VarCopy( arr_data, argv[1] ) ;
	    }
	    break;
	    case ARRAY_Float32:
	    {
		arr_ptr = IDL_MakeTempArray( IDL_TYP_FLOAT, (int)num_dims,
					     (IDL_MEMINT *)dims,
					     IDL_BARR_INI_NOP, &arr_data ) ;
		if( get_data )
		{
		    memcpy( arr_data->value.arr->data,
		            arr_vals, n_elements * 4 ) ;
		}
		else 
		{
		    memset( arr_data->value.arr->data, 0, n_elements * 4 ) ;
		}

		IDL_VarCopy( arr_data, argv[1] ) ;
	    }
	    break;
	    case ARRAY_Float64:
	    {
		arr_ptr = IDL_MakeTempArray( IDL_TYP_DOUBLE, (int)num_dims,
					     (IDL_MEMINT *)dims,
					     IDL_BARR_INI_NOP, &arr_data ) ;
		if( get_data )
		{
		    memcpy( arr_data->value.arr->data,
		            arr_vals, n_elements * 8 ) ;
		}
		else 
		{
		    memset( arr_data->value.arr->data, 0, n_elements * 8 ) ;
		}

		IDL_VarCopy( arr_data, argv[1] ) ;
	    }
	    break;
	    case ARRAY_String:
	    case ARRAY_URL:
	    {
		IDL_STRING *str_arr = 0 ;
		str_arr = (IDL_STRING *)IDL_MakeTempArray( IDL_TYP_STRING,
		                                           (int)num_dims,
							   (IDL_MEMINT *)dims,
							   IDL_BARR_INI_NOP,
							   &arr_data ) ;
		NodeContent_String **strings = (NodeContent_String **)arr_vals ;
		unsigned int i_elements = 0 ;
		for( i_elements = 0; i_elements < n_elements; i_elements++ )
		{
		    unsigned int slen = 0 ;
		    char *tmpstr = 0 ;
		    if( get_data )
		    {
			NodeContent_String *snode =strings[i_elements] ;
			slen = snode->iLength + 1 ;
			tmpstr = (char *)malloc( slen ) ;
			strcpy( tmpstr, snode->sContent ) ;
		    }
		    else
		    {
			slen = strlen( "" ) + 1 ;
			tmpstr = (char *)malloc( slen ) ;
			strcpy( tmpstr, "" ) ;
		    }
		    IDL_StrStore( &(str_arr[i_elements]), tmpstr ) ;
		    free( tmpstr ) ;
		    str_arr[i_elements].slen = slen ;
		}
		IDL_VarCopy( arr_data, argv[1] ) ;
	    }
	    break;
	    default:
	    {
		return lReturn ;
	    }
	    break;
	}
    }

    lReturn->value.ul = 1ul ;

    return lReturn ;
}

// opendap_get_grid()
// Get the data from a DDS grid variable.
//    argv[0] - (Input ) content pointer.
//    argv[1-12] - (Output) Named IDL variable to receive the maps.
//    argv[13]  (Output) Named IDL variable to receive map names
//    argv[14]  (Input ) 1: get the data
//    returns -          Status (1=ok, 0=fail).
//
IDL_VPTR
opendap_get_grid( int argc, IDL_VPTR argv[] )
{
    IDL_VPTR lReturn ;
    lReturn = IDL_Gettmp() ; // Allocate memory for return variable.
    lReturn->type = IDL_TYP_ULONG ;
    lReturn->value.ul = 0ul ;

    void *vcontent = (void *)argv[0]->value.memint ;
    if( !vcontent )
    {
	return lReturn ;
    }

    NodeContent_Grid *content = (NodeContent_Grid *)vcontent ;

    char *arr_ptr = 0 ;
    IDL_VPTR arr_data ;
    IDL_MEMINT dims[12] ;

    unsigned long get_data = argv[14]->value.ul ;
    unsigned int n_maps = content->ctMappings ;
    unsigned int i_maps = 1 ;
    unsigned int i_argv = 1 ;

    IDL_VPTR dim_names ;
    int dim_names_sizes[1] ;
    dim_names_sizes[0] = n_maps ;
    unsigned int name_len = 0 ;
    IDL_STRING *str_arr = 0 ;
    str_arr = (IDL_STRING *)IDL_MakeTempArray( IDL_TYP_STRING, (int)1, 
					       (IDL_MEMINT *)&dim_names_sizes, 
					       IDL_BARR_INI_NOP, &dim_names ) ;

    NodeContent_Array *gmap = 0 ;

    for( ; i_maps <= n_maps; i_maps++, i_argv++ )
    {
	gmap = content->maps1[i_maps] ;

	char *gmap_name = gmap->sArrayName ;
	name_len = 0 ;
	if( gmap_name ) name_len = strlen( gmap_name ) + 1 ;
	if( name_len > 64 ) name_len = 64 ;
	char *tmpstr = 0 ;
	// Handle case where dimension is not a named dimension;
	// Create a container for the dimension with a 'space' char in it.
	if( name_len == 0 )
	{
	    name_len = strlen( "" ) + 1 ;
	    tmpstr = (char *)malloc( name_len ) ;
	    strcpy( tmpstr, "" ) ;
	}
	else
	{
	    tmpstr = (char *)malloc( name_len ) ;
	    strcpy( tmpstr, gmap_name ) ;
	}
	IDL_StrStore( &(str_arr[i_maps-1]), tmpstr ) ;
	str_arr[i_maps-1].slen = name_len ;

	eARRAY_TYPE array_type = gmap->eType ;
	void *arr_vals = gmap->array ;
	unsigned int num_dims = gmap->ctDimensions ;
	unsigned int i_num_dims = 0 ;
	unsigned int n_elements = 1;

	for( i_num_dims = 0; i_num_dims < num_dims; i_num_dims++ )
	{
	    dims[num_dims - i_num_dims - 1] =
		    gmap->iDimensionLengthArray1[i_num_dims+1] ;
	    n_elements *= gmap->iDimensionLengthArray1[i_num_dims+1] ;
	}

	switch( array_type )
	{
	    case ARRAY_Byte:
	    {
		arr_ptr = IDL_MakeTempArray( IDL_TYP_BYTE, (int)num_dims, 
					     (IDL_MEMINT *)dims,
					     IDL_BARR_INI_NOP,
					     &arr_data ) ;
		if( get_data )
		{
		    memcpy( arr_data->value.arr->data,
		            arr_vals, n_elements * 1 ) ;
		}
		else 
		{
		    memset( arr_data->value.arr->data, 0, n_elements * 1 ) ;
		}

		IDL_VarCopy(arr_data, argv[i_argv]);
	    }
	    break;

	    case ARRAY_Int16:
	    {
		arr_ptr = IDL_MakeTempArray( IDL_TYP_INT, (int)num_dims,
					     (IDL_MEMINT *)dims,
					     IDL_BARR_INI_NOP,
					     &arr_data ) ;
		if( get_data )
		{
		    memcpy( arr_data->value.arr->data,
		            arr_vals, n_elements * 2 ) ;
		}
		else 
		{
		    memset( arr_data->value.arr->data, 0, n_elements * 2 ) ;
		}

		IDL_VarCopy(arr_data, argv[i_argv]);
	    }
	    break;
	    case ARRAY_UInt16:
	    {
		arr_ptr = IDL_MakeTempArray( IDL_TYP_UINT, (int)num_dims,
					     (IDL_MEMINT *)dims,
					     IDL_BARR_INI_NOP, &arr_data ) ;

		if( get_data )
		{
		    memcpy( arr_data->value.arr->data,
		            arr_vals, n_elements * 2 ) ;
		}
		else 
		{
		    memset( arr_data->value.arr->data, 0, n_elements * 2 ) ;
		}

		IDL_VarCopy(arr_data, argv[i_argv]);
	    }
	    break;
	    case ARRAY_Int32:
	    {
		arr_ptr = IDL_MakeTempArray( IDL_TYP_LONG, (int)num_dims,
					     (IDL_MEMINT *)dims,
					     IDL_BARR_INI_NOP, &arr_data ) ;
		if( get_data )
		{
		    memcpy( arr_data->value.arr->data,
		            arr_vals, n_elements * 4 ) ;
		}
		else 
		{
		    memset( arr_data->value.arr->data, 0, n_elements * 4 ) ;
		}

		IDL_VarCopy( arr_data, argv[i_argv] ) ;
	    }
	    break;
	    case ARRAY_UInt32:
	    {
		arr_ptr = IDL_MakeTempArray( IDL_TYP_ULONG, (int)num_dims,
					     (IDL_MEMINT *)dims,
					     IDL_BARR_INI_NOP, &arr_data ) ;
		if( get_data )
		{
		    memcpy( arr_data->value.arr->data,
		            arr_vals, n_elements * 4 ) ;
		}
		else 
		{
		    memset( arr_data->value.arr->data, 0, n_elements * 4 ) ;
		}

		IDL_VarCopy( arr_data, argv[i_argv] ) ;
	    }
	    break;
	    case ARRAY_Float32:
	    {
		arr_ptr = IDL_MakeTempArray( IDL_TYP_FLOAT, (int)num_dims,
					     (IDL_MEMINT *)dims,
					     IDL_BARR_INI_NOP, &arr_data ) ;
		if( get_data )
		{
		    memcpy( arr_data->value.arr->data,
		            arr_vals, n_elements * 4 ) ;
		}
		else 
		{
		    memset( arr_data->value.arr->data, 0, n_elements * 4 ) ;
		}

		IDL_VarCopy( arr_data, argv[i_argv] ) ;
	    }
	    break;
	    case ARRAY_Float64:
	    {
		arr_ptr = IDL_MakeTempArray( IDL_TYP_DOUBLE, (int)num_dims,
					     (IDL_MEMINT *)dims,
					     IDL_BARR_INI_NOP, &arr_data ) ;
		if( get_data )
		{
		    memcpy( arr_data->value.arr->data,
		            arr_vals, n_elements * 8 ) ;
		}
		else 
		{
		    memset( arr_data->value.arr->data, 0, n_elements * 8 ) ;
		}

		IDL_VarCopy( arr_data, argv[i_argv] ) ;
	    }
	    break;
	    default:
	    {
		return lReturn ;
	    }
	    break;
	}
    }

    IDL_VarCopy( dim_names, argv[13] ) ;

    lReturn->value.ul = 1ul ;

    return lReturn ;
}

// opendap_get_grid_values()
// Get the values (not the maps) of the specified grid
//    argv[0] - (Input ) content pointer.
//    argv[1] - (Output) Named IDL variable to receive values
//
void
opendap_get_grid_values( int argc, IDL_VPTR argv[] )
{
    IDL_VPTR vcontent ;
    vcontent = IDL_Gettmp() ;
    vcontent->type = IDL_TYP_MEMINT ;
    vcontent->value.memint = (IDL_MEMINT)0UL ;

    void *content = (void *) argv[0]->value.memint ;
    if( !content )
    {
	IDL_VarCopy( vcontent, argv[1] ) ;
	return ;
    }
    NodeContent_Grid *gcontent = (NodeContent_Grid *)content ;
    vcontent->value.memint = (IDL_MEMINT)gcontent->values ;
    IDL_VarCopy( vcontent, argv[1] ) ;
}

// opendap_get_sequence_info()
// Get the values (not the maps) of the specified grid
//    argv[0] - (Input ) content pointer.
//    argv[1] - (Output) Named IDL variable to receive num_rows
//    argv[2] - (Output) Named IDL variable to receive num_fields
//
void
opendap_get_sequence_info( int argc, IDL_VPTR argv[] )
{
    void *vcontent = (void *)argv[0]->value.memint ;
    NodeContent_Sequence *content = (NodeContent_Sequence *)vcontent ;

    IDL_VPTR num_rows ;
    num_rows = IDL_Gettmp() ;
    num_rows->type = IDL_TYP_ULONG ;
    num_rows->value.ul = (IDL_ULONG)content->ctRows ;
    IDL_VarCopy( num_rows, argv[1] ) ;

    IDL_VPTR num_fields ;
    num_fields = IDL_Gettmp() ;
    num_fields->type = IDL_TYP_ULONG ;
    num_fields->value.ul = (IDL_ULONG)content->ctFields ;
    IDL_VarCopy( num_fields, argv[2] ) ;
}

// opendap_get_sequence_field()
// Get the values (not the maps) of the specified grid
//    argv[0] - (Input ) node pointer.
//    argv[1] - (Input ) content pointer.
//    argv[2] - (Input ) field number
//    argv[3] - (Output) Named IDL variable to receive field name
//    argv[4] - (Output) Named IDL variable to receive field type
//    argv[5] - (Output) Named IDL variable to receive field sub node
//    argv[6] - (Output) Named IDL variable to receive field content pointer
//    returns - Status (1=ok, 0=fail)
//
IDL_VPTR
opendap_get_sequence_field( int argc, IDL_VPTR argv[] )
{
    IDL_VPTR lReturn ;
    lReturn = IDL_Gettmp() ; // Allocate memory for return variable.
    lReturn->type = IDL_TYP_ULONG ;
    lReturn->value.ul = 0ul ;

    DAPNode *node = (DAPNode *)argv[0]->value.memint ;
    void *vcontent = (void *)argv[1]->value.memint ;
    NodeContent_Sequence *sequence = (NodeContent_Sequence *)vcontent ;
    unsigned long i_field = argv[2]->value.ul ;

    unsigned int num_fields = sequence->ctFields ;
    unsigned int num_subnodes = node->iSubNodeCount ;

    if( num_fields != num_subnodes )
    {
	return lReturn ;
    }
    if( i_field < 0 || i_field > num_fields )
    {
	return lReturn ;
    }

    DAPNode *subnode = node->nodeSub[i_field] ;

    char *field_name = subnode->sName ;
    IDL_VPTR name_arr ;
    name_arr = IDL_Gettmp() ;
    name_arr->type = IDL_TYP_STRING ;
    name_arr->value.str.slen = strlen( field_name ) ;
    name_arr->value.str.stype = 0 ;
    name_arr->value.str.s = (char *)field_name ;
    IDL_VarCopy( name_arr, argv[3] ) ;

    IDL_VPTR field_type ;
    field_type = IDL_Gettmp() ;
    field_type->type = IDL_TYP_ULONG ;
    field_type->value.ul = (IDL_ULONG)0 ;
    field_type->value.ul = (IDL_ULONG)sequence->aeFieldType[i_field] ;
    IDL_VarCopy( field_type, argv[4] ) ;

    IDL_VPTR field_node ;
    field_node = IDL_Gettmp() ;
    field_node->type = IDL_TYP_MEMINT ;
    field_node->value.memint = (IDL_MEMINT)subnode ;
    IDL_VarCopy( field_node, argv[5] ) ;

    void **field_values = sequence->aFieldValue ;
    IDL_VPTR field_content ;
    field_content = IDL_Gettmp() ;
    field_content->type = IDL_TYP_MEMINT ;
    if( field_values )
    {
	field_content->value.memint = (IDL_MEMINT)field_values[i_field] ;
    }
    else
    {
	field_content->value.memint = (IDL_MEMINT)0 ;
    }
    IDL_VarCopy( field_content, argv[6] ) ;
    
    lReturn->value.ul = 1ul ;

    return lReturn ;
}

// opendap_get_sequence_row()
// Get the pointer from the sequence aFieldValue[] for the given row
//    argv[0] - (Input ) node pointer.
//    argv[1] - (Input ) content pointer.
//    argv[2] - (Input ) field type
//    argv[3] - (Input ) row number
//    argv[4] - (Output) Named IDL variable to receive row value ptr
//    returns - Status (1=ok, 0=fail)
//
IDL_VPTR
opendap_get_sequence_row( int argc, IDL_VPTR argv[] )
{
    IDL_VPTR lReturn ;
    lReturn = IDL_Gettmp() ; // Allocate memory for return variable.
    lReturn->type = IDL_TYP_ULONG ;
    lReturn->value.ul = 0ul ;

    DAPNode *node = (DAPNode *)argv[0]->value.memint ;
    void *field_content = (void *)argv[1]->value.memint ;
    unsigned int field_type = argv[2]->value.ul ;
    unsigned int row_num = argv[3]->value.ul ;

    IDL_VPTR row_content ;
    row_content = IDL_Gettmp() ;
    row_content->type = IDL_TYP_MEMINT ;

    switch( field_type )
    {
	case NODE_Byte:
	{
	    if( !field_content )
		row_content->value.memint = (IDL_MEMINT)0 ;
	    else
	    {
		unsigned int offset = row_num * 1 ;
		row_content->value.memint = (IDL_MEMINT)field_content+offset ;
	    }
	}
	break ;
	case NODE_Int16:
	case NODE_UInt16:
	{
	    if( !field_content )
		row_content->value.memint = (IDL_MEMINT)0 ;
	    else
	    {
		unsigned int offset = row_num * 2 ;
		row_content->value.memint = (IDL_MEMINT)field_content+offset ;
	    }
	}
	break ;
	case NODE_Int32:
	case NODE_UInt32:
	case NODE_Float32:
	{
	    if( !field_content )
		row_content->value.memint = (IDL_MEMINT)0 ;
	    else
	    {
		unsigned int offset = row_num * 4 ;
		row_content->value.memint = (IDL_MEMINT)field_content+offset ;
	    }
	}
	break ;
	case NODE_Float64:
	{
	    if( !field_content )
		row_content->value.memint = (IDL_MEMINT)0 ;
	    else
	    {
		unsigned int offset = row_num * 8 ;
		row_content->value.memint = (IDL_MEMINT)field_content+offset ;
	    }
	}
	break ;
	case NODE_String:
	case NODE_URL:
	{
	    if( !field_content )
		row_content->value.memint = (IDL_MEMINT)node->nodeContent ;
	    else
	    {
		NodeContent_String **scontent = (NodeContent_String **)field_content ;
		row_content->value.memint = (IDL_MEMINT)scontent[row_num] ;
	    }
	}
	break ;
	case NODE_Sequence:
	{
	    if( !field_content )
		row_content->value.memint = (IDL_MEMINT)node->nodeContent ;
	    else
	    {
		NodeContent_Sequence **scontent = (NodeContent_Sequence **)field_content ;
		row_content->value.memint = (IDL_MEMINT)scontent[row_num] ;
	    }
	}
	break ;
	case NODE_Array:
	{
	    if( !field_content )
		row_content->value.memint = (IDL_MEMINT)node->nodeContent ;
	    else
	    {
		NodeContent_Array **acontent = (NodeContent_Array **)field_content ;
		row_content->value.memint = (IDL_MEMINT)acontent[row_num] ;
	    }
	}
	break ;
	case NODE_Grid:
	{
	    if( !field_content )
		row_content->value.memint = (IDL_MEMINT)node->nodeContent ;
	    else
	    {
		NodeContent_Grid **gcontent = (NodeContent_Grid **)field_content ;
		row_content->value.memint = (IDL_MEMINT)gcontent[row_num] ;
	    }
	}
	break ;
	case NODE_Structure:
	{
	    if( !field_content )
		row_content->value.memint = (IDL_MEMINT)node->nodeContent ;
	    else
	    {
		NodeContent_Structure **scontent = (NodeContent_Structure **)field_content ;
		row_content->value.memint = (IDL_MEMINT)scontent[row_num] ;
	    }
	}
	break ;
	default:
	{
	    return lReturn ;
	}
	break ;
    }
    IDL_VarCopy( row_content, argv[4] ) ;

    lReturn->value.ul = 1ul ;

    return lReturn ;
}

// opendap_get_struct_info()
// Get the information about a structure
//    argv[0] - (Input ) content pointer.
//    argv[1] - (Output) Named IDL variable to receive num_members
//
void
opendap_get_struct_info( int argc, IDL_VPTR argv[] )
{
    void *vcontent = (void *)argv[0]->value.memint ;
    NodeContent_Structure *content = (NodeContent_Structure *)vcontent ;

    IDL_VPTR num_members ;
    num_members = IDL_Gettmp() ;
    num_members->type = IDL_TYP_ULONG ;
    num_members->value.ul = (IDL_ULONG)content->ctMembers ;
    IDL_VarCopy( num_members, argv[1] ) ;
}

// opendap_get_struct_member()
// Get the content for the specified member of a NodeContent_Structure
//    argv[0] - (Input ) content pointer.
//    argv[1] - (Input ) member number to get
//    argv[2] - (Output) Named IDL variable to receive member type
//    argv[3] - (Output) Named IDL variable to receive member content ptr
//
IDL_VPTR
opendap_get_struct_member( int argc, IDL_VPTR argv[] )
{
    IDL_VPTR lReturn ;
    lReturn = IDL_Gettmp() ; // Allocate memory for return variable.
    lReturn->type = IDL_TYP_ULONG ;
    lReturn->value.ul = 0ul ;

    IDL_VPTR member ;
    member = IDL_Gettmp() ;
    member->type = IDL_TYP_MEMINT ;
    member->value.memint = (IDL_MEMINT)0UL ;

    void *vcontent = (void *)argv[0]->value.memint ;
    NodeContent_Structure *content = (NodeContent_Structure *)vcontent ;

    unsigned int mem_num = argv[1]->value.ul ;
    if( mem_num < 0 || mem_num > content->ctMembers )
    {
	printf( "%s: %d: Structure member number is out of range\n",
	        __FILE__, __LINE__ ) ;
	IDL_VarCopy( member, argv[3] ) ;
	return lReturn ;
    }

    IDL_VPTR num_members ;
    num_members = IDL_Gettmp() ;
    num_members->type = IDL_TYP_ULONG ;
    num_members->value.ul = (IDL_ULONG)content->array_eNodeType[ mem_num ] ;
    IDL_VarCopy( num_members, argv[2] ) ;

    member->value.memint = (IDL_MEMINT)content->array_NodeContent[ mem_num ] ; 
    IDL_VarCopy( member, argv[3] ) ;

    lReturn->value.ul = 1ul ;

    return lReturn ;
}

